local tfm = {}

tfm.ChangeExtension = function(path, newext)
-- Changes the extension that ends the specified file path with a new one.
-- If no new extension is specified, it deletes the current one.
	newext = newext or ""
	if newext:len() > 0 then
		if newext:sub(1, 1) ~= "." then
			newext = table.concat({".", newext})
		end
	end
	local dotindex = path:find("%.[^\\%.]*$")
	if dotindex == nil then
		return table.concat({path, newext})
	end
	return table.concat({path:sub(1, dotindex-1), newext})
end

tfm.CombinePath = function(pathsarray)
-- Combines an array of path parts in a complete path. Basically you don't need to worry about missing or redundant slashes.
	local result = ""
	for _,p in ipairs(pathsarray) do
		result = table.concat({result, "\\", p})
	end
	result = result:sub(2):gsub("\\+", "\\")
	return result
end

tfm.ExtractFileName = function(path, noext)
-- Extracts the file name from the specified file path.
-- If noext is set to true, the extension is removed.
	local result = path:match("[^\\]+$")
	if noext then
		return tfm.ChangeExtension(result)
	end
	return result
end

tfm.ReadLOL = function(sourcename)
-- Loads a compiled Lua file and returns the module it generates.
	local io = require "io"
	local lolfile = assert(io.open(sourcename, "rb"))
	local rawdata = lolfile:read("*all")
	local data = table.concat({rawdata:sub(1, 5), "\0", rawdata:sub(7)})
	local lol = assert(loadstring(data))
	local oldT = _G.T
	_G.T = function(s) return s end -- Temporarily insert dummy T function
	lol("tempmodule")
	_G.T = oldT
	return tempmodule
end

tfm.GetPlainLOL = function(lolmodule)
-- Returns a string containing a code that should provide the same result as the original module.
	if type(lolmodule) ~= "table" then
		return ""
	end
	local moduletypes = {"card", "cards", "card_pool", "deck"}
	local moduletype = nil
	for _,m in ipairs(moduletypes) do
		if lolmodule[m] then
			moduletype = m
			break
		end
	end
	if moduletype == nil then
		return ""
	end
	local result = (require "dump").tostring(lolmodule[moduletype])
	result = result:gsub("^{", "module(..., package.seeall)\n\n"..moduletype.." = {", 1)
	if moduletype == "card" or moduletype == "cards" then
		result = result:gsub('data = "(.-)",\n(\t*)language_code', 'data = T("%1"),\n%2language_code')

		-- Make code human-friendly
		result = result:gsub('%scode = ".-[^\\]",\n', function(m)
			local unbackslashed = function(s)
				local ch = {
					["\\a"] = '\\007', --'\a' alarm             Ctrl+G BEL
					["\\b"] = '\\008', --'\b' backspace         Ctrl+H BS
					["\\f"] = '\\012', --'\f' formfeed          Ctrl+L FF
					["\\n"] = '\\010', --'\n' newline           Ctrl+J LF
					["\\r"] = '\\013', --'\r' carriage return   Ctrl+M CR
					["\\t"] = '\\009', --'\t' horizontal tab    Ctrl+I HT
					["\\v"] = '\\011', --'\v' vertical tab      Ctrl+K VT
					["\\\n"] = '\\010',--     newline
					["\\\\"] = '\\092',--     backslash
					["\\'"] = '\\039', --     apostrophe
					['\\"'] = '\\034', --     quote
				}
				return s:gsub("(\\.)", ch):gsub("\\(%d%d?%d?)", function(n)
					return string.char(tonumber(n))
				end)
			end
			return unbackslashed(m:gsub('code = "(.+)",\n', 'code = [[\n%1]],\n'))
		end)

	end
	return result
end

tfm.CompileLOL = function(sourcename, destname)
-- Compiles the source file into a DotP-compatible destination file.
-- If no destination file is specified, it will create a file with lol extension in the same folder as the source.
	local io = require "io"
	local luafile = assert(io.open(sourcename, "rb"))
	local plaindata = luafile:read("*all")
	local startindex = assert(plaindata:find("module"))
	local rawdata = string.dump(assert(loadstring(plaindata:sub(startindex))))
	local data = table.concat({rawdata:sub(1, 5), string.char(95), rawdata:sub(7)})
	destname = destname or tfm.ChangeExtension(sourcename, "lol")
	local lolfile = assert(io.open(destname, "wb"))
	assert(lolfile:write(data))
	assert(lolfile:close())
end

tfm.DecompileLOL = function(sourcename, destname)
-- Automatically runs ReadLOL and GetPlainLOL on a source file, creating a "decompiled" destination file.
-- If no destination file is specified, it will create a file with lua extension in the same folder as the source.
	local lol = tfm.ReadLOL(sourcename)
	local str = tfm.GetPlainLOL(lol)
	destname = destname or tfm.ChangeExtension(sourcename, "lua")
	local io = require "io"
	local luafile = assert(io.open(destname, "w"))
	assert(luafile:write(str))
	assert(luafile:close())
end

return tfm